/// tkopengl.cpp
///
/// (c) 2001-2006 by Bastian Spiegel <bs@tkscript.de>
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#include "static_tkopengl.h"

#define TKS_NUM_TMP_VECTOR   32
#define TKS_NUM_TMP_IVECTOR  32
#define TKS_NUM_TMP_VECTOR4  32

#define TKS_MASK_TMP_VECTOR  31
#define TKS_MASK_TMP_IVECTOR 31
#define TKS_MASK_TMP_VECTOR4 31

#ifdef DX_STATIC_TKOPENGL
#define YAC_NO_EXPORTS defined 
#define YAC_CUST_OBJECT defined 
#define YAC_CUST_STRING defined 
#define YAC_CUST_EVENT defined 
#define YAC_CUST_VALUE defined 
#define YAC_CUST_LISTNODE defined 
#define YAC_CUST_TREENODE defined 
#define YAC_CUST_STREAMBASE defined 
#define YAC_CUST_BUFFER defined 
#define YAC_CUST_INTARRAY defined 
#define YAC_CUST_FLOATARRAY defined 
#define YAC_CUST_NUMBEROBJECTS defined 
#endif

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define YAC_PRINTF defined
#define YAC_BIGSTRING
#include <yac.h>

#undef INADDR_ANY
#undef INADDR_NONE
#ifndef DX_STATIC_TKOPENGL
#include <tks-list.h>
#include <yac_host.cpp>
#endif
#include <tks-opengl.h>
#include <GL/glu.h>
#include "bitmap.h"
#include "image.h"
#include "tkstexture.h"
#include "chaosmath.h"
#include "tksmath.h"
//#include "tksmathstatic.h"
#include "color.h"
#include "tksopengl.h"

extern sUI clid_Color;
extern sUI clid_IVector;
extern sUI clid_Matrix;
extern sUI clid_Texture;
extern sUI clid_Vector;
extern sUI clid_Vector4;
extern sUI clid_VectorArray;

#ifndef DX_STATIC_TKOPENGL
YAC_Host *yac_host;
#endif



// ---- to emulate the (old) ivector(), vector(), vector4() expressions using functions
// ---- ...these were faster once because of constant expr optimizations...
static sUI      tmp_next_ivector;
static _IVector*tmp_ivector;
static sUI      tmp_next_vector;
static _Vector* tmp_vector;
static sUI      tmp_next_vector4;
static _Vector4*tmp_vector4;
//static _IVector tmp_ivector[TKS_NUM_TMP_IVECTOR];
//static _Vector  tmp_vector[TKS_NUM_TMP_VECTOR];
//static _Vector4 tmp_vector4[TKS_NUM_TMP_VECTOR4];

sUI YAC_CALL _zglReadColorPixel(sSI _x, sSI _y) {
	sSI pixel=0;
	::glReadPixels(_x, _y, 1, 1, GL_RGBA, GL_UNSIGNED_BYTE, (void*)&pixel);
	return pixel;
}

sF32 YAC_CALL _zglReadDepthPixel(sSI _x, sSI _y) {
	sF32 pixel=0;
	::glReadPixels(_x, _y, 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, (void*)&pixel);
	return pixel;
}

void YAC_CALL _zglClipPlane(sSI _i, sF32 _eq0, sF32 _eq1, sF32 _eq2, sF32 _eq3) {
	static sF64 eq[4];
   eq[0] = _eq0;
   eq[1] = _eq1;
   eq[2] = _eq2;
   eq[3] = _eq3;
   ::glClipPlane(_i, eq);
}

YAC_Object *YAC_CALL _ivector(sSI _x, sSI _y) {
	_IVector *v=&tmp_ivector[tmp_next_ivector++];
	tmp_next_ivector&=TKS_MASK_TMP_IVECTOR;
	v->x=_x; v->y=_y;
	return v;
}

YAC_Object *YAC_CALL _vector(sF32 _x, sF32 _y, sF32 _z) {
	_Vector *v=&tmp_vector[tmp_next_vector++];
	tmp_next_vector&=TKS_MASK_TMP_VECTOR;
	v->x=_x; v->y=_y; v->z=_z;
	return v;
}

YAC_Object *YAC_CALL _vector4(sF32 _x, sF32 _y, sF32 _z, sF32 _w) {
	_Vector4 *v=&tmp_vector4[tmp_next_vector4++];
	tmp_next_vector4&=TKS_MASK_TMP_VECTOR4;
	v->x=_x; v->y=_y; v->z=_z; v->w=_w;
	return v;
}

sUI YAC_CALL _mathC32Multiply(sUI c32a, sUI c32b, sF32 _bscale) {
	sUI r32=0;
	sU8 ca;
	sU8 cb;
	// ---- alpha ----
	ca=(c32a>>24)&0xFF;
	cb=(c32b>>24)&0xFF;
	r32|=((sUI)(ca+ (ca*cb*1.0f/256.0f-ca)*_bscale))<<24;
	// ---- red ----
	ca=(c32a>>16)&0xFF;
	cb=(c32b>>16)&0xFF;
	r32|=((sUI)(ca+ (ca*cb*1.0f/256.0f-ca)*_bscale))<<16;
	// ---- green ----
	ca=(c32a>>8)&0xFF;
	cb=(c32b>>8)&0xFF;
	r32|=((sUI)(ca+ (ca*cb*1.0f/256.0f-ca)*_bscale))<<8;
	// ---- green ----
	ca=c32a&0xFF;
	cb=c32b&0xFF;
	r32|=((sUI)(ca+ (ca*cb*1.0f/256.0f-ca)*_bscale));

	return r32;
}

sUI YAC_CALL _mathC32Blend(sUI c32a, sUI c32b, sF32 _bscale) {
	sU32 r32=0;
	sU8 ca;
	sU8 cb;
	// ---- alpha ----
	ca=(c32a>>24)&0xFF;
	cb=(c32b>>24)&0xFF;
	r32|=((sU32)(ca+ (cb-ca)*_bscale))<<24;
	// ---- red ----
	ca=(c32a>>16)&0xFF;
	cb=(c32b>>16)&0xFF;
	r32|=((sU32)(ca+ (cb-ca)*_bscale))<<16;
	// ---- green ----
	ca=(c32a>>8)&0xFF;
	cb=(c32b>>8)&0xFF;
	r32|=((sU32)(ca+ (cb-ca)*_bscale))<<8;
	// ---- green ----
	ca=c32a&0xFF;
	cb=c32b&0xFF;
	r32|=((sU32)(ca+ (cb-ca)*_bscale));

	return r32;
}

void YAC_CALL _mathNormal(YAC_Object *_v1, YAC_Object *_v2, YAC_Object*_v3, YAC_Object*_ret){
    if(YAC_CHK(_v1, clid_Vector))
        if(YAC_CHK(_v2, clid_Vector))
            if(YAC_CHK(_v3, clid_Vector))
				if(YAC_CHK(_ret, clid_Vector))
				{
					_Vector *v1=(_Vector*)_v1;
					_Vector *v2=(_Vector*)_v2;
					_Vector *v3=(_Vector*)_v3;
					_Vector *r=(_Vector*)_ret;
					sFVector a;
					sFVector b;
					a.x= v1->x - v2->x;
		            a.y= v1->y - v2->y;
	                a.z= v1->z - v2->z;                   
					b.x= v2->x - v3->x;
					b.y= v2->y - v3->y;
					b.z= v2->z - v3->z;
					r->x=a.y*b.z - a.z*b.y;
					r->y=a.z*b.x - a.x*b.z;
					r->z=a.x*b.y - a.y*b.x;
				}
}

void YAC_CALL _mathUnproject(YAC_Object *_v, YAC_Object *_rnear, YAC_Object *_rfar) {
	if(YAC_CHK(_v, clid_Vector))
        if(YAC_CHK(_rnear, clid_Vector))
            if(YAC_CHK(_rfar, clid_Vector))
            {
                GLdouble mvmatrix[4*4];
                GLdouble prmatrix[4*4];
                ::glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
                ::glGetDoublev(GL_PROJECTION_MATRIX, prmatrix);
                _Vector *vnear=(_Vector*)_rnear;
                _Vector *vfar=(_Vector*)_rfar;
                _Vector *v=(_Vector*)_v;
                GLdouble x=(GLdouble)v->x; 
                GLint view[4];
                ::glGetIntegerv(GL_VIEWPORT, view);
                GLdouble y=(GLdouble)view[3]/*tks_HAL->view_sx??*/-v->y;
                GLdouble rx,ry,rz;
                ::gluUnProject(x, y, 0, mvmatrix, prmatrix, view, &rx, &ry, &rz);
                vnear->x=(sF32)rx;
                vnear->y=(sF32)ry;
                vnear->z=(sF32)rz;
                ::gluUnProject(x, y, 1, mvmatrix, prmatrix, view, &rx,&ry,&rz);
                vfar->x=(sF32)rx;
                vfar->y=(sF32)ry;
                vfar->z=(sF32)rz;
            }
}

void YAC_CALL _mathProject(YAC_Object *_v, YAC_Object *_vret) {
	if(YAC_CHK(_v, clid_Vector))
        if(YAC_CHK(_vret, clid_Vector))
        {
            _Vector *v=(_Vector *)_v;
            _Vector *vret=(_Vector *)_vret;
		
			GLdouble prjmatrix[4*4];
			GLdouble mvmatrix[4*4];
			GLint view[4];
			::glGetDoublev(GL_PROJECTION_MATRIX, prjmatrix);
			::glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
			::glGetIntegerv(GL_VIEWPORT, view);
			// project mmm
			GLdouble rx;
			GLdouble ry;
			GLdouble rz;
			::gluProject((GLdouble)v->x, (GLdouble)v->y, (GLdouble)v->z, mvmatrix, prjmatrix, view, &rx, &ry, &rz);
			vret->x=(sF32)rx;
			vret->y=(sF32)ry;
			vret->z=(sF32)rz;
        }
}

void YAC_CALL _mathReflect(YAC_Object *_in, YAC_Object *_n, YAC_Object *_ret) {
	if(YAC_CHK(_in, clid_Vector))
		if(YAC_CHK(_n, clid_Vector))
			if(YAC_CHK(_ret, clid_Vector))
			{
				_Vector *vi=(_Vector*)_in;
				_Vector *n=(_Vector*)_n;
				_Vector *r=(_Vector*)_ret;
				// ---- calc reflection vector ----
				// ---- I: incoming ray        ----
				// ---- S: surface normal      ----
				// ---- L= unit(I)             ----
				// ---- N= unit(S)             ----
				// ---- L'=(L*N)*N             ----
				// ---- R'=-L + 2*L'           ----
				// ---- R= R'*|I|              ----
				// ---- R: reflected ray       ----
				sFVector vt;
				sFVector vt1;
				sFVector vt2;
				vt2.x=n->x; vt2.y=n->y; vt2.z=n->z;
				vt2.unit();                             // ---- N= unit(S)
				vt1.x=vi->x; vt1.y=vi->y; vt1.z=vi->z;  // incoming ray
				sF32 l=vt1.Abs();                       // ---- |I|           
				vt1.unit();                             // ---- L= unit(I)
				// ---- L'=(L*N)*N; ----
				sF32 f_2=vt1.Dot(vt2);  
				vt.x=vt2.x; vt.y=vt2.y; vt.z=vt2.z;
				vt.scalef(f_2);
				r->x=vt.x; r->y=vt.y; r->z=vt.z;
				r->Add(*r);
				r->Sub(vt1);
				// ---- R= R'*|M|; ----
				r->scalef(l);
			}
}

sF32 YAC_CALL _mathDistancePointPlane(YAC_Object *_p, YAC_Object *_q, YAC_Object *_n) {
	if(YAC_CHK(_p, clid_Vector))
		if(YAC_CHK(_q, clid_Vector))
			if(YAC_CHK(_n, clid_Vector))
			{
				_Vector *p=(_Vector*)_p;
				_Vector *q=(_Vector*)_q;
				_Vector *n=(_Vector*)_n;
				sFVector v;
				v.x=p->x-q->x; v.y=p->y-q->y; v.z=p->z-q->z;
				return (v.Dot(*n)/n->Abs());
			}
	return 0.0f;
}

void YAC_CALL _mathVectorMin(YAC_Object *_v1, YAC_Object *_v2, YAC_Object *_r) {
	if(YAC_CHK(_r, clid_Vector))
		if(YAC_CHK(_v1, clid_Vector))
			if(YAC_CHK(_v2, clid_Vector))
			{
				_Vector *v1=(_Vector*)_v1;
				_Vector *v2=(_Vector*)_v2;
				_Vector *r=(_Vector*)_r;
				r->x=v1->x<v2->x?v1->x:v2->x;
				r->y=v1->y<v2->y?v1->y:v2->y;
				r->z=v1->z<v2->z?v1->z:v2->z;
			}
}

void YAC_CALL _mathVectorMax(YAC_Object *_v1, YAC_Object *_v2, YAC_Object *_r) {
	if(YAC_CHK(_r, clid_Vector))
		if(YAC_CHK(_v1, clid_Vector))
			if(YAC_CHK(_v2, clid_Vector))
			{
				_Vector *v1=(_Vector*)_v1;
				_Vector *v2=(_Vector*)_v2;
				_Vector *r=(_Vector*)_r;
				r->x=v1->x>=v2->x?v1->x:v2->x;
				r->y=v1->y>=v2->y?v1->y:v2->y;
				r->z=v1->z>=v2->z?v1->z:v2->z;
			}
}

sUI YAC_CALL _mathC32ToC8(sUI _c32) {
	return (
		((((_c32>>16)&0xFF)*80)>>8) + 
		(((_c32>> 8 )&0xFF)>>1) + 
		((((_c32    )&0xFF)*47)>>8)
		);
}

void YAC_CALL _zglColor1f(sF32 _cf) {
	::glColor4f(_cf, _cf, _cf, _cf);
}


/// ---------------------------------------------------------------------------
#include "ying_Color.cpp"
#include "ying_Color_Color.cpp"
#include "ying_Math.cpp"
#include "ying_Math_IVector.cpp"
#include "ying_Math_Math.cpp"
#include "ying_Math_Matrix.cpp"
#include "ying_Math_Vector.cpp"
#include "ying_Math_Vector4.cpp"
#include "ying_Math_VectorArray.cpp"
#include "ying_Math_WrappedFloat.cpp"
#include "ying_OpenGL.cpp"
#include "ying_OpenGL_DisplayList.cpp"
#include "ying_Texture.cpp"
#include "ying_Texture_Texture.cpp"


// ---------------------------------------------------------------------------
#ifdef DX_STATIC_TKOPENGL
void YAC_CALL YAC_Init_tkopengl(YAC_Host *_host) {
#else
void YAC_CALL YAC_Init(YAC_Host *_host) {
#endif
	if(_host->yacGetDebugLevel())
		_host->printf("tkopengl::YAC_Init called.\n");
	yac_host=_host;

	YAC_Init_Math    (_host);
	YAC_Init_Color   (_host);
	YAC_Init_OpenGL  (_host);
	YAC_Init_Texture (_host);

	// ---- volatile vectors ----
	tmp_next_ivector=0;
	tmp_next_vector=0;
	tmp_next_vector4=0;

   tmp_ivector = new _IVector[TKS_NUM_TMP_IVECTOR];
   tmp_vector  = new _Vector [TKS_NUM_TMP_VECTOR ];
   tmp_vector4 = new _Vector4[TKS_NUM_TMP_VECTOR4];

	sUI i; 
	for(i=0; i<TKS_NUM_TMP_IVECTOR; i++) tmp_ivector[i].class_ID=clid_IVector;
	for(i=0; i<TKS_NUM_TMP_VECTOR;  i++) tmp_vector [i].class_ID=clid_Vector;
	for(i=0; i<TKS_NUM_TMP_VECTOR4; i++) tmp_vector4[i].class_ID=clid_Vector4;

	if(yac_host->yacGetDebugLevel())
		yac_host->yacPrint("tkopengl::YAC_Init() finished.\n");
}

#ifdef DX_STATIC_TKOPENGL
void YAC_CALL YAC_Exit_tkopengl(YAC_Host *_host) {
#else
void YAC_CALL YAC_Exit(YAC_Host *_host) {
#endif

   delete [] tmp_ivector; tmp_ivector = 0;
   delete [] tmp_vector;  tmp_vector  = 0;
   delete [] tmp_vector4; tmp_vector4 = 0;

	YAC_Exit_Texture (_host);
	YAC_Exit_OpenGL  (_host);
	YAC_Exit_Color   (_host);
	YAC_Exit_Math    (_host);

	if(yac_host->yacGetDebugLevel())
		yac_host->yacPrint("tkopengl::YAC_Exit() finished.\n");
}

#ifndef DX_STATIC_TKOPENGL
sUI YAC_CALL YAC_Version(void) {
	return 0x00080006;
}
#endif // not DX_STATIC_TKOPENGL

