/// tkstexture.h
///
/// (c) 1999-2004 Bastian Spiegel <bs@tkscript.de>
///     - published under terms of the GNU general public license (GPL)
///
///
#ifndef __TKSTEXTURE_H__
#define __TKSTEXTURE_H__

YG("Texture");


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//~                                                                ~
//~ Texture                                                        ~
//~                                                                ~
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
YC class _Texture : public YAC_Object {
public:
#define TEX_16BIT              YCI (1<<0)
#define TEX_REPEAT_S           YCI (1<<1)
#define TEX_REPEAT_T           YCI (1<<2)
#define TEX_DELETE             YCI (1<<3)
#define TEX_AUTOMIPMAP         YCI (1<<4)
#define TEX_MODULATE           YCI (1<<5)
#define TEX_BLEND              YCI (1<<6)
#define TEX_DECAL              YCI (1<<7)
#define TEX_REPLACE            YCI (1<<8)
#define TEX_MINFILTERLINEAR    YCI (1<<9)
#define TEX_MINFILTERTRILINEAR YCI (1<<10)
#define TEX_MAGFILTERLINEAR    YCI (1<<11)
#define TEX_MAGFILTERTRILINEAR YCI (1<<12)
#define TEX_ALPHA              YCI (1<<13)
#define TEX_UPLOADED           YCI (1<<14)
#define TEX_1D                 YCI (1<<15)
#define TEX_FORCEHOSTBYTEORDER YCI (1<<16)
#define TEX_INTENSITY          YCI (1<<17)
#define TEX_REPEATMASK         YCI (TEX_REPEAT_S|TEX_REPEAT_T)
#define TEX_ENVMASK            YCI (TEX_MODULATE|TEX_BLEND|TEX_DECAL|TEX_REPLACE)
#define TEX_FILTERMASK         YCI (TEX_MINFILTERLINEAR|TEX_MINFILTERTRILINEAR|TEX_MAGFILTERLINEAR|TEX_MAGFILTERTRILINEAR)

public:
	Image image;
	sUI   gl_nameid;
	sUI   flags;
	sF32  priority;

public:
	 _Texture(void);
	 ~_Texture();
	 YAC(_Texture);

     static void Expand2n(sSI *_sx, sSI *_sy);
 
	 sUI   YAC_CALL yacArrayAlloc(sUI _num, sUI,sUI,sUI);
	 sUI   YAC_CALL yacArrayGetWidth(void);
	 sUI   YAC_CALL yacArrayGetHeight(void);
	 sUI   YAC_CALL yacArrayGetElementType(void);
	 sUI   YAC_CALL yacArrayGetElementByteSize(void);
	 sUI   YAC_CALL yacArrayGetStride(void);
	 void *YAC_CALL yacArrayGetPointer(void);
	 void  YAC_CALL yacArraySet(void *_context, sUI _index, YAC_Value *_value);
    void  YAC_CALL yacArrayGet(void *_context, sUI _index, YAC_Value *_ret);

	 void upload(sBool _newnameid, sUI _x, sUI _y, sUI _w, sUI _h);
	 void simpleBlur            (Image *_t, Image *_j);
	 void simpleBlurX           (Image *_t, Image *_j);
	 void simpleBlurY           (Image *_t, Image *_j);

	 YM void _setFlags             (sSI _flags);
	 YM sUI  _getFlags             (void);
	 YM sSI  _loadImage            (YAC_String *, sSI _scx, sSI _scy, sSI _z);
	 YM sSI  _saveImage            (YAC_String *);
	 YM void _upload               (void);
	 YM void _setScreen            (YAC_Object *_srctex);
	 YM sSI  _alloc                (sSI _sx, sSI _sy, sSI _z);
	 YM void _scaleFrom            (YAC_Object *);
	 YM void _bind                 (void);
	 YM void _free                 (void);
	 YM void _freeImage            (void);
	 YM void _setScreenArea        (YAC_Object *_srctex, YAC_Object *_lt, YAC_Object *_sz);
	 YM sF32 _setPriority          (sF32);
	 YM void _setEnvColor          (sSI);
 	 YM void _copyRegion           (YAC_Object *_texsource, YAC_Object *_ivecp, YAC_Object *_ivecs, YAC_Object *_ivecd);
 	 YM void _clear                (sSI _clearcolor);
	 YM void _unload               (void);
     YM sSI  _getSx                (void);
     YM sSI  _getSy                (void);
     YM sSI  _getZ                 (void);
	 YM sSI  _loadLocalImage       (YAC_String *_s, sSI _scx, sSI _scy, sSI _z);
	 YM void _drawPixels           (YAC_Object *_rasterposvector);
    YM void _drawPixels2i         (sSI _x, sSI _y);
    YM void _drawPixels2f         (sF32 _x, sF32 _y);
    YM void _drawPixels3f         (sF32 _x, sF32 _y, sF32 _z);
	 YM void _swapScreen           (YAC_Object *_tex);
	 YM void _setFgcolor           (sSI);
	 YM sSI  _getFgcolor           (void);
	 YM sSI  _getId                (void);
	 YM void _applyColorMap        (YAC_Object *_t1, YAC_Object *_cm);
	 YM void _colorKeyToAlpha      (sSI _ckey);
	 YM sSI  _grayscale            (void);
	 YM void _interleaveAlpha      (YAC_Object *_alphatex);
	 YM void _adjustGamma          (sF32);
	 YM void _adjustBrightness     (sF32);
 	 YM void _blend                (YAC_Object *_srca, YAC_Object *_srcb, sSI _ratio255);
    YM sF32 _countPixels          (sSI _colorkey);
    YM void _multiply             (YAC_Object *_stexa, YAC_Object *_stexb);
	 YM void _setXY32              (sSI _x, sSI _y, sSI _c32);
	 YM sUI  _getXY32              (sSI _x, sSI _y);
    YM void _drawPolygon          (YAC_Object *_intarray, sSI _color);
    YM void _drawLine             (YAC_Object *_q, YAC_Object *_p, sSI _color);
    YM void _drawFilledBox        (YAC_Object *_p, YAC_Object *_s, sSI _color);
	 YM void _simpleBlur           (YAC_Object *_j);
	 YM void _simpleBlurX          (YAC_Object *_j);
	 YM void _simpleBlurY          (YAC_Object *_j);
    YM void _scroll               (sSI _dx, sSI _dy);
    YM void _expand2n             (void);
	 YM sSI  _alloc2n              (sSI _sx, sSI _sy, sSI _z);
	 YM void _createMask           (sSI _ckey);
	 YM sUI  _getStride            (void);
	 YM void _update               (void);
    YM void _updateRegion         (sSI _x, sSI _y, sSI _w, sSI _h);
    YM void _renderFixedText      (YAC_Object *_fontTex, sSI _charWidth, YAC_Object *_text, sSI _x, sSI _y, YAC_Object *_colorIntArray, YAC_Object *_attrIntArray);
};

#endif
